from .adjustments import ADJUSTMENT_MENUITEM
from .assets import EQUITYCASHPOSITION_MENUITEM
from .claim import (
    CONSOLIDATED_TRADE_SUMMARY_TABLE_MENUITEM,
    CLAIM_MENUITEM,
    NEGATIVEACCOUNTPRODUCT_MENUITEM,
    PNL_MENUITEM,
    AUM_NNM_PROGRESSION_MENUITEM,
)
from .custodians import CustodianMenuItem
from .fees import FEES_MENUITEM, GAIN_MENUITEM
from .instrument_prices import VALUATIONDATAAUM_MENUITEM
from .portfolios import ModelPortfolioMenuItem, PortfolioMenuItem
from .positions import (
    AGGREGATED_ASSETPOSITION_LIQUIDITY_MENUITEM,
    ASSETPOSITION_MENUITEM,
)
from .product_groups import PRODUCTGROUP_MENUITEM
from .product_performance import (
    PRODUCTPERFORMANCECOMPARISONLIST_MENUITEM,
    PRODUCTPERFORMANCELIST_MENUITEM,
    PRODUCTPERFORMANCENNMLIST_MENUITEM,
)
from .products import PRODUCT_MENUITEM, PRODUCTCUSTOMER_MENUITEM
from .registers import REGISTER_MENUITEM
from .roles import PORTFOLIOROLE_MENUITEM
from .trades import SUBSCRIPTION_REDEMPTION_MENUITEM, TRADE_MENUITEM
from .portfolio_cash_flow import PORTFOLIO_DAILY_CASH_FLOW
from .reconciliations import ACCOUNT_RECONCILIATION_MENU_ITEM
from .orders import OrderProposalMenuItem



from datetime import date

import numpy as np
import pandas as pd
from django.db.models import QuerySet
from wbfdm.contrib.metric.backends.base import AbstractBackend, Metric
from wbfdm.contrib.metric.exceptions import MetricInvalidParameterError

from wbportfolio.models import AssetPosition, Portfolio


class PortfolioMetricBaseBackend(AbstractBackend[Portfolio]):
    BASKET_MODEL_CLASS = Portfolio

    def convert_df_into_metrics(
        self,
        df: pd.DataFrame | pd.Series,
        key: str,
        portfolio_id: int,
        metric_date: date,
        columns_map: dict[int | str, str] | None = None,
        dependency_metrics: list["Metric"] | None = None,
    ) -> list[Metric]:
        if not dependency_metrics:
            dependency_metrics = []
        metrics = []
        df = df.replace([np.inf, -np.inf, np.nan], None)
        if isinstance(df, pd.DataFrame):
            if columns_map:
                df = df.rename(columns=columns_map)
            for instrument_id, metric in df.to_dict("index").items():
                metrics.append(
                    Metric(
                        basket_id=portfolio_id,
                        basket_content_type_id=self.content_type.id,
                        key=key,
                        metrics=metric,
                        date=metric_date,
                        instrument_id=instrument_id,
                        dependency_metrics=dependency_metrics,
                    )
                )
        else:
            if columns_map:
                df = df.rename(index=columns_map)
            metrics.append(
                Metric(
                    basket_id=portfolio_id,
                    basket_content_type_id=self.content_type.id,
                    key=key,
                    metrics=df.to_dict(),
                    date=metric_date,
                    dependency_metrics=dependency_metrics,
                )
            )
        return metrics

    def _get_valid_date(self, portfolio):
        qs = portfolio.assets.filter(is_estimated=False)
        if self.val_date is not None:
            qs = qs.filter(date__lte=self.val_date)
        try:
            return qs.latest("date").date
        except AssetPosition.DoesNotExist:
            raise MetricInvalidParameterError() from None

    def get_queryset(self) -> QuerySet[Portfolio]:
        product_portfolios = super().get_queryset().filter_active_and_tracked()
        try:
            last_position_date = (
                AssetPosition.objects.filter(portfolio__in=product_portfolios, is_estimated=False).latest("date").date
            )
        except AssetPosition.DoesNotExist:
            last_position_date = date.today()
        return product_portfolios.filter_invested_at_date(last_position_date)



import codecs
import csv
import datetime

import pandas as pd
from wbcore.contrib.currency.models import Currency

from wbportfolio.import_export.utils import convert_string_to_number


def parse(import_source):
    # Load file into a CSV DictReader and convert the encoding to latin1 due to hyphonation
    csv_file = import_source.file.open()
    csv_file = codecs.iterdecode(csv_file, "latin1")

    # Read file into a CSV Dict Reader
    csv_reader = csv.DictReader(csv_file, delimiter=";")

    # Iterate through the CSV File and parse the data into a list
    data = list()
    for basket in csv_reader:
        if basket["Assetart"] == "CPON" or (
            basket["Assetart"] == "TRES" and "Sichtguthaben" not in basket["Assetbezeichnung"]
        ):
            continue
        identifier = basket["Fonds-Nr."].strip()
        valuation_date = datetime.datetime.strptime(basket["Datum"], "%Y%m%d")

        currency_key = basket["Währung"]
        currency = Currency.objects.get(key=currency_key)
        group_currency = Currency.objects.get(key=basket["Assetwährung"])

        if basket["Assetart"] == "TRES":
            initial_currency_fx_rate = convert_string_to_number(basket["Wertpapierkurs"])

            ticker = "CASH"
            title = currency.title
            initial_price = 1
            underlying_quote = {
                "instrument_type": "cash",
                "ticker": ticker,
                "name": title,
                "currency__key": currency.key,
            }
        elif basket["Assetart"] == "VMOB":
            instrument_type = "equity"
            if "STRUKT" in basket.get("TL2", ""):
                instrument_type = "product"
            initial_currency_fx_rate = float(currency.convert(valuation_date, group_currency))
            initial_price = basket["Wertpapierkurs"]

            exchange = f'X{basket["Bloomberg Code"]}' if basket["Bloomberg Code"] else None
            underlying_quote = {
                "instrument_type": instrument_type,
                "isin": basket["ISIN"],
                "exchange": {"mic_code": exchange},
                "name": basket["Assetbezeichnung"],
                "currency__key": currency.key,
            }

        initial_shares = basket["Stk / Nominal"]

        data.append(
            {
                "underlying_quote": underlying_quote,
                "exchange": underlying_quote.get("exchange", None),
                "portfolio": {
                    "instrument_type": "product_group",
                    "identifier": identifier,
                },
                "is_estimated": False,
                "currency__key": currency_key,
                "date": valuation_date.strftime("%Y-%m-%d"),
                "asset_valuation_date": valuation_date.strftime("%Y-%m-%d"),
                "initial_price": round(convert_string_to_number(initial_price), 4),
                "initial_currency_fx_rate": initial_currency_fx_rate,
                "initial_shares": round(convert_string_to_number(initial_shares), 4),
            }
        )
    csv_file.close()

    df = pd.DataFrame(data)  # quick Hack to compute weighting on import
    df["weighting"] = df.initial_currency_fx_rate * df.initial_price * df.initial_shares
    df["weighting"] = df.weighting / df.weighting.sum()

    return {"data": df.to_dict("records")}



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PerformancePandasEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-list", request=self.request)


class PerformanceComparisonEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbportfolio:product-list", request=self.request)


class ProductPerformanceNetNewMoneyEndpointConfig(PerformancePandasEndpointConfig):
    pass



from wbcore.menus import ItemPermission, MenuItem

PRODUCTGROUP_MENUITEM = MenuItem(
    label="Product Groups",
    endpoint="wbportfolio:product_group-list",
    permission=ItemPermission(
        method=lambda request: request.user.is_internal, permissions=["wbportfolio.view_productgroup"]
    ),
)



from datetime import date

from django.db.models import QuerySet
from wbfdm.models import Classification, Instrument, InstrumentList


class InstrumentMixin:
    trade_date: date

    def get_instruments(self, **kwargs):
        return self._filter_instruments(self._get_instruments(**kwargs), **kwargs)

    def _get_instruments(
        self,
        classification_ids: list[int] | None = None,
        instrument_ids: list[int] | None = None,
        instrument_list_id: int | None = None,
        **kwargs,
    ) -> QuerySet[Instrument]:
        """
        Use the provided kwargs to return a list of instruments as universe.
        - If classifications are given, we returns all the instrument linked to these classifications
        - Or directly from a static list of instrument ids
        - fallback to the last effective portfolio underlying instruments list
        """
        if not instrument_ids:
            instrument_ids = []
        if classification_ids:
            for classification in Classification.objects.filter(id__in=classification_ids):
                for children in classification.get_descendants(include_self=True):
                    for company in children.instruments.all():
                        instrument_ids.append(company.get_primary_quote().id)
        if instrument_list_id:
            instrument_list = InstrumentList.objects.get(id=instrument_list_id)
            instrument_ids.extend(list(instrument_list.get_instruments(self.trade_date).values_list("id", flat=True)))

        if not instrument_ids:
            instrument_ids = list(self.effective_portfolio.positions_map.keys())

        return Instrument.objects.filter(id__in=instrument_ids)

    def _filter_instruments(
        self, instruments: QuerySet[Instrument], instrument_filter_kwargs: dict[str, str] | None = None, **kwargs
    ) -> QuerySet[Instrument]:
        instruments = instruments.filter(is_cash=False)
        if instrument_filter_kwargs:
            instruments = instruments.filter(**instrument_filter_kwargs)
        return instruments



import pytest
from faker import Faker
from psycopg.types.range import NumericRange
from wbcompliance.factories.risk_management import RuleThresholdFactory

from wbportfolio.risk_management.backends.liquidity_risk import (
    RuleBackend as LiquidityRiskRuleBackend,
)

fake = Faker()


@pytest.mark.django_db
class TestProductRuleModel:
    @pytest.fixture
    def liquidity_risk_backend(
        self,
        weekday,
        instrument,
    ):
        return LiquidityRiskRuleBackend(
            weekday,
            instrument,
            json_parameters={"liquidation_factor": 3, "redemption_pct": 0.80},
            thresholds=[
                RuleThresholdFactory.create(range=NumericRange(lower=5, upper=None)),
            ],
        )

    def test_check_rule_product_liquidity(
        self, weekday, instrument, instrument_price_factory, asset_position_factory, liquidity_risk_backend
    ):
        res = list(liquidity_risk_backend.check_rule())  # no position, no risk incident
        assert len(res) == 0

        shares = 100

        volume_50d = (shares * liquidity_risk_backend.redemption_pct * liquidity_risk_backend.liquidation_factor) / 6
        instrument_price_factory.create(date=weekday, instrument=instrument, volume_50d=volume_50d, calculated=False)
        asset_position_factory.create(
            date=weekday, underlying_instrument=instrument, is_estimated=False, initial_shares=shares
        )

        res = list(liquidity_risk_backend.check_rule())  # no position, no risk incident
        assert len(res) == 1
        incident = res[0]
        assert incident.report_details["Total Shares"] == shares
        assert incident.report_details["Volume 50D"] == volume_50d



from django.db.models import TextChoices

class ExecutionStatus(TextChoices):
    PENDING = "PENDING", "Pending"
    IN_DRAFT = "IN_DRAFT", "In Draft"
    COMPLETED = "COMPLETED", "Completed"
    CANCELLED = "CANCELLED", "Cancelled"
    REJECTED = "REJECTED", "Rejected"
    FAILED = "FAILED", "Failed"
    UNKNOWN = "UNKNOWN", "Unknown"



class ExecutionInstruction(TextChoices):

    MARKET_ON_CLOSE = "MARKET_ON_CLOSE", "Market On Close" # no parameter
    GUARANTEED_MARKET_ON_CLOSE = "GUARANTEED_MARKET_ON_CLOSE", "Guaranteed Market On Close" # no parameter
    GUARANTEED_MARKET_ON_OPEN  = "GUARANTEED_MARKET_ON_OPEN", "Guaranteed Market On Open" # no parameter
    GPW_MARKET_ON_CLOSE = "GPW_MARKET_ON_CLOSE", "GPW Market On Close" # no parameter
    MARKET_ON_OPEN = "MARKET_ON_OPEN", "Market On Open" # no parameter
    IN_LINE_WITH_VOLUME = "IN_LINE_WITH_VOLUME", "In Line With Volume" # 1 parameter "Percentage"
    LIMIT_ORDER = "LIMIT_ORDER", "Limit Order" # 2 parameters "limit and cutoff"
    VWAP = "VWAP", "VWAP" # 2 parameters
    TWAP = "TWAP", "TWAP" # 2 paramters



class RoutingException(Exception):
    def __init__(self, errors):
        # messages: a list of strings
        super().__init__()  # You can pass a summary to the base Exception
        self.errors = errors

    def __str__(self):
        return str(self.errors)
